from operator import attrgetter
import os
import sys

from testtools import try_imports

get_platform = try_imports(['sysconfig.get_platform', 'distutils.util.get_platform'])

from pkgme.project_info import (
    MultipleExternalHelpersInfo,
    SingleExternalHelperInfo,
    )
from pkgme.run_script import run_script, ScriptMissing


def get_backend_dir(underunder_file, backend):
    backend_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(underunder_file), os.pardir,
            'backends', backend))
    # When 'python setup.py test' is run with virtualenv, backend_dir will not
    # point to the right place.  It will point into the
    # build/lib.{platform}-{version} directory and that cannot be used as a
    # landmark to find the backends directory.  Assuming we're running the
    # tests from the top of the tree, we do the best we can to find the actual
    # landmark.
    parts = backend_dir.split(os.sep)
    try:
        i = parts.index('build')
        # Stolen from site.py's addbuilddir().
        s = 'lib.%s-%.3s' % (get_platform(), sys.version)
        if hasattr(sys, 'gettotalrefcount'):
            s += '-pydebug'
        if parts[i + 1] == s:
            del parts[i:i+2]
        backend_dir = os.sep.join(parts)
    except (ValueError, IndexError):
        # We're (probably) not in the build environment.
        pass
    return backend_dir


# Where do the backends live?  The environment variable overrides the system
# built-in and discovered development locations.
PKGME_BACKEND_PATHS = os.environ.get('PKGME_BACKEND_PATHS')
if PKGME_BACKEND_PATHS is None:
    EXTERNAL_BACKEND_PATHS = [os.path.join(sys.prefix, "share", "pkgme", "backends")]

    root_dir = os.path.dirname(get_backend_dir(__file__, ''))

    if os.path.exists(os.path.join(root_dir, 'setup.py')):
        EXTERNAL_BACKEND_PATHS.insert(
            0, os.path.join(root_dir, 'pkgme', 'backends'))
else:
    EXTERNAL_BACKEND_PATHS = PKGME_BACKEND_PATHS.split(os.pathsep)


def get_info_for(path, loader=None, selector_cls=None):
    if loader is None:
        loader = get_default_loader()
    backends = loader.load()
    if selector_cls is None:
        selector_cls = get_default_selector
    selector = selector_cls(backends)
    return selector.get_info(path)


def _external_helpers_loader_factory():
    return ExternalHelpersBackendLoader(EXTERNAL_BACKEND_PATHS)


_default_loader_factory = _external_helpers_loader_factory


def get_default_loader():
    global _default_loader_factory
    return _default_loader_factory()


def set_default_loader_factory(loader_factory):
    global _default_loader_factory
    _default_loader_factory = loader_factory


def reset_default_loader_factory():
    global _default_loader_factory
    _default_loader_factory = _external_helpers_loader_factory


def get_default_selector(backends):
    return BackendSelector(backends)


class Backend(object):

    def __init__(self, name):
        self.name = name

    def want(self, path):
        raise NotImplementedError(self.want)

    def get_info(self, path):
        raise NotImplementedError(self.get_info)


class ExternalHelpersBackend(Backend):

    WANT_SCRIPT_NAME = "want"

    def __init__(self, name, basepath):
        super(ExternalHelpersBackend, self).__init__(name)
        self.basepath = basepath

    def want(self, path):
        try:
            out = run_script(self.basepath, self.WANT_SCRIPT_NAME, path)
        except ScriptMissing:
            raise AssertionError(
                "Backend %s (%s) has no '%s' script"
                % (self.name, self.basepath, self.WANT_SCRIPT_NAME))
        result = out.strip()
        try:
            score = int(result)
        except ValueError:
            raise AssertionError(
                "Backend %s (%s) returned non-integer score from "
                "'%s' script: %s"
                % (self.name, self.basepath,
                    self.WANT_SCRIPT_NAME, str(result)))
        return score

    def get_info(self, path):
        cls = MultipleExternalHelpersInfo
        if os.path.exists(
            os.path.join(
                self.basepath, SingleExternalHelperInfo.INFO_SCRIPT_NAME)):
            cls = SingleExternalHelperInfo
        return cls(self.basepath, path)


class StaticBackend(Backend):

    def __init__(self, name, score, info, expected_path=None):
        super(StaticBackend, self).__init__(name)
        self.score = score
        self.info = info
        self.expected_path = expected_path

    def want(self, path):
        if self.expected_path is not None:
            assert self.expected_path == path, (
                "Called with cwd of %s, expected %s"
                % (path, self.expected_path))
        return self.score

    def get_info(self, path):
        if self.expected_path is not None:
            assert self.expected_path == path, (
                "Called with cwd of %s, expected %s"
                % (path, self.expected_path))
        return self.info


class NoBackend(Exception):
    pass


class BackendSelector(object):

    def __init__(self, backends):
        self.backends = backends

    def get_info(self, path):
        high_score = 0
        eligble = []
        for backend in self.backends:
            score = backend.want(path)
            if score > high_score:
                eligble = [backend]
                high_score = score
            elif score == high_score:
                eligble.append(backend)
        if len(eligble) < 1 or high_score < 1:
            # FIXME: some info about the problem
            raise NoBackend()
        backend = eligble[0]
        if len(eligble) > 1:
            eligble.sort(key=attrgetter('name'))
            backend = eligble[0]
        return backend.get_info(path)


class BackendLoader(object):

    def load(self):
        raise NotImplementedError(self.load)


class ExternalHelpersBackendLoader(BackendLoader):

    def __init__(self, paths):
        self.paths = paths

    def load(self):
        backends = []
        names = set()
        for path in self.paths:
            path = os.path.abspath(path)
            if not os.path.isdir(path):
                continue
            fnames = os.listdir(path)
            for fname in fnames:
                if fname.startswith("."):
                    continue
                if fname in names:
                    continue
                backend_path = os.path.join(path, fname)
                if not os.path.isdir(backend_path):
                    continue
                backends.append(
                    ExternalHelpersBackend(fname, backend_path))
                names.add(fname)
        return backends


class StaticLoader(BackendLoader):

    def __init__(self, backends):
        self.backends = backends

    def load(self):
        return self.backends
