"""Things to help package binary tarballs.

HIGHLY EXPERIMENTAL. USE AT YOUR OWN PERIL.

At the moment, we are assuming a great many things about these binary
tarballs.

* That they represent some sort of application to be run from an Ubuntu
  desktop

* Specifically, that although they might have many executables, only one is
  the "main executable" mentioned in the 'desktop' file

* They are written in C or C++ and that all dependencies can be determined
  by inspecting the included object files

* That the entire contents of the tarball can be copied into
  /opt/<package-name> and run from there

* That we have a metadata file, called 'devportal-metadata.json', in JSON,
  that specifies:
  * package_name

The expectation is that this metadata file is generated from the developer
portal.
"""

# XXX: No idea about architecture

__all__ = [
    'get_install_file',
    'guess_executable',
    'iter_executables',
    'METADATA_FILE',
   ]


import os
import subprocess

from pkgme.info_elements import (
    ApplicationName,
    Categories,
    Executable,
    PackageName,
    TagLine,
    )
from pkgme.package_files import (
    DEBIAN_DIR,
    Desktop,
    )
from pkgme.project_info import DictInfo


METADATA_FILE = 'devportal-metadata.json'

# Keys found in the metadata file.
CATEGORIES = 'categories'
# It's called package_name in the database, so this probably makes sense.
PACKAGE_NAME = 'package_name'
TAGLINE = 'tagline'


# Always depend on these.
#
# XXX: jml just doing this because iamfuzz does. Need to either experiment or
# consult expert.
DEPENDS = '${shlibs:Depends}, ${misc:Depends}'

# XXX: No idea about how icons will be there. Ignoring for now.

def guess_executable(package_name, executables):
    """
    From a list of executables, guess which one is likely to be the main
    executable.
    """

    # The lower the score, the more likely it is to be the right executable.

    def score_name(name):
        # Names that match exactly are good.
        return -int(name.lower() == package_name.lower())

    def score_path(executable):
        # The deeper the path, the less likely it is to be the one.
        return executable.count('/')

    def rank(executable):
        name = os.path.basename(executable)
        # The alpha-sorting of the base name is a tie-breaker.
        return score_name(name), score_path(executable), name

    try:
        return sorted(executables, key=rank)[0]
    except IndexError:
        return None


def iter_executables(path):
    """Iterate through all executable files under 'path'.

    Paths yielded will be relative to 'path'. Directories will *not* be
    yielded. "Executable" is determined by filesystem permissions.
    """
    for root, dirs, files in os.walk(path):
        for filename in files:
            file_path = os.path.join(root, filename)
            if os.access(file_path, os.X_OK):
                yield os.path.relpath(file_path, path)


def get_install_file(package_name, path, include_desktop=False):
    """Generate the install file for 'package_name'."""
    lines = []
    # Sorting not actually needed for functionality, but makes the tests more
    # reliable.
    for filename in sorted(os.listdir(path)):
        if filename in (DEBIAN_DIR, METADATA_FILE):
            # We don't want to install the 'debian/' directory or the metadata
            # file.
            continue
        lines.append('%s opt/%s' % (filename, package_name))
    if include_desktop:
        lines.append(
            'debian/%s.desktop usr/share/application' % (package_name,))
    # Ending the file with a newline is basic good manners.
    lines.append('')
    return '\n'.join(lines)


def get_desktop_file(package_name, path, tagline=None, categories=None):
    executable = guess_executable(package_name, iter_executables(path))
    info = {
        PackageName.name: package_name,
        ApplicationName.name: package_name.capitalize(),
        Executable.name: '/opt/%s/%s' % (package_name, executable),
        TagLine.name: tagline,
        Categories.name: categories,
        }
    return Desktop.from_info(DictInfo(info))


def get_file_type(path):
    return get_file_types([path])[0]


def get_file_types(paths):
    paths = list(paths)
    if not paths:
        return []
    cmd = ['file', '-b'] + paths
    return subprocess.Popen(
        cmd, stdout=subprocess.PIPE).communicate()[0].splitlines()


def ldd(binary_paths):
    # ldd has output of the form:
    #
    # path/to/elf/object:
    #    linux-vdso.so.1 =>  (0x00007fff4a7ff000)
    #    libc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007f7472623000)
    #    /lib64/ld-linux-x86-64.so.2 (0x00007f74729e5000)
    #
    # i.e.
    #    <name> => <path> (<address>)
    #
    # where <name> and <path> are both optional. If <name> is not given, then
    # there is no arrow (=>).
    binary_paths = list(binary_paths)
    if not binary_paths:
        return {}
    cmd = ['ldd'] + binary_paths
    output = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
    libraries = {}
    current = []
    if len(binary_paths) == 1:
        libraries[binary_paths[0]] = current
    for line in output.splitlines():
        if line.endswith(':'):
            current.sort()
            current = libraries[line.strip(':')] = []
            continue
        # Last token is always irrelevant hex address.
        tokens = line.split()[:-1]
        name, path = None, None
        if '=>' in tokens:
            name = tokens[0]
            if len(tokens) > 2:
                path = tokens[2]
        else:
            path = tokens[0]
        current.append((name, path))
    current.sort()
    return libraries


def get_shared_library_dependencies(paths):
    # XXX: Not sure which of these is more useful. Specifically, if the
    # library isn't installed on the system, then I don't know whether it will
    # appear in ldd as the library_path. -- jml
    libraries = ldd(paths)
    for libs in libraries.values():
        for name, library_path in libs:
            if library_path:
                yield library_path


def find_shared_library_dependencies(path):
    return set(get_shared_library_dependencies(iter_binaries(path)))


def iter_binaries(path):

    def is_binary(file_type):
        return file_type.startswith('ELF')

    for root, dirs, files in os.walk(path):
        paths = [os.path.join(root, filename) for filename in files]
        types = get_file_types(paths)
        for file_path, file_type in zip(paths, types):
            if is_binary(file_type):
                yield file_path


def apt_file(filename):
    # XXX: Assumes apt-file update has been run.

    # XXX: Passing the full file name since it seems the least likely to yield
    # multiple dependencies.  If 'get_shared_library_dependencies' ought to
    # return just the names, then this will have to change to be a regex, and
    # we'll need to think about ambiguity resolution much sooner.
    # cmd = ['apt-file', 'find', '-x', '/%s$' % (filename,)]

    # XXX: Currently matching badly due to architecture.
    cmd = ['apt-file', 'find', '-F', filename]
    output = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
    for line in output.splitlines():
        yield line.split(': ')[0]


def guess_dependencies(path):
    deps = set()
    # XXX: ldd probably is recursive. This means it finds *all* the libraries
    # that are needed. However, we only need to specify the top-level
    # libraries (e.g. libc6 should almost never be a dependency).
    for library in find_shared_library_dependencies(path):
        for dep in apt_file(library):
            deps.add(dep)
    return deps
