import json


class MissingInfoError(Exception):
    pass


class InfoElement(object):

    required = False
    default = None

    @property
    def name(self):
        raise NotImplementedError(
            "Define a name attribute on %s" % self.__class__)

    @property
    def description(self):
        raise NotImplementedError(
            "Define a name attribute on %s" % self.__class__)

    @classmethod
    def get_value(cls, info):
        vals = info.get_all([cls.name])
        val = vals.get(cls.name, None)
        if val is None:
            if cls.required:
                raise MissingInfoError(
                    "The backend didn't provide a value for '%s'" % cls.name)
            val = cls.default
        val = cls.clean(val)
        return val

    @classmethod
    def clean(cls, value):
        """Validate and modify the value as needed.

        This method should do any validation needed on the value. It can also
        modify the value if needed. It should return the new value or the
        passed value if there is nothing to do.
        """
        return value


class Buildsystem(InfoElement):

    name = "buildsystem"
    description = "The debhelper buildsystem to use to build the package."


class PackageName(InfoElement):

    name = "package_name"
    description = "The name of the package."
    required = True


class Section(InfoElement):

    name = "section"
    description = "The section the source package should go in."
    default = "misc"


class Maintainer(InfoElement):

    name = "maintainer"
    description = """The maintainer of the package.

    Should be in the form::

        Name <email>

    """
    default = "A Person <someone@example.com>"


class Homepage(InfoElement):

    name = "homepage"
    description = """The homepage for the project.

    The URL of the website for the project.
    """


class BuildDepends(InfoElement):

    name = "build_depends"
    description = """The build dependencies of the package."""

    @classmethod
    def clean(cls, value):
        if value is None:
            parts = []
        elif isinstance(value, (str, unicode)):
            parts = [value]
        else:
            parts = value
        # FIXME: should replace if the dependency isn't strict
        # enough
        # FIXME: check is too lax, needs to parse and compare package
        # names
        if not filter(lambda x: "debhelper" in x, parts):
            parts += ["debhelper (>= 7)"]
        return ", ".join(parts)


class DebhelperAddons(InfoElement):

    name = "debhelper_addons"
    description = """The debhelper addons to use."""


class Architecture(InfoElement):

    name = "architecture"
    description = """The architecture(s) to build the binary package for.

    This can be one of the following.

    any
        to build on every architecture (arch-dependent).
    all
        to build once and use the result on all architecture
        (arch-independent).
    a list of architectures (arch-dependent)
        to build on specific architectures only.

    """
    default = "any"


class Depends(InfoElement):

    name = "depends"
    description = """The dependencies of the binary package."""


class Description(InfoElement):

    name = "description"
    description = "The description of the package."
    default = "a package"


class Version(InfoElement):

    name = "version"
    description = "The version of the project."
    default = 0


class ExtraFiles(InfoElement):

    name = 'extra_files'
    description = """Any extra files needed to build the package.

    This is specified as a dictionary mapping filename to file contents.  The
    filename must be relative to the directory being packaged.
    """
    default = {}

    @classmethod
    def clean(cls, value):
        if isinstance(value, dict):
            return value
        return json.loads(value)


# The following are for 'binary'.

class ApplicationName(InfoElement):

    name = 'application_name'
    description = "The name of an application."


class Categories(InfoElement):

    name = 'categories'
    description = "The categories the application belongs to."


class Executable(InfoElement):

    name = 'executable'
    description = "The path to the executable for an application."


class TagLine(InfoElement):

    name = 'tagline'
    description = "A short line describing the application."
