import datetime
import os

from debian import changelog

from pkgme.info_elements import (
    ApplicationName,
    Architecture,
    BuildDepends,
    Buildsystem,
    Categories,
    DebhelperAddons,
    Depends,
    Description,
    Executable,
    ExtraFiles,
    Homepage,
    Maintainer,
    PackageName,
    Section,
    TagLine,
    Version,
    )
from pkgme.project_info import DictInfo
from pkgme.template_file import TemplateFile


DEBIAN_DIR = "debian"


class BasicFile(object):

    def __init__(self, path, content, overwrite=True, elements=None):
        self.path = path
        self._content = content
        self.overwrite = overwrite
        if elements is None:
            elements = []
        self.elements = elements

    def get_contents(self):
        return self._content


class PackageFile(object):

    overwrite = True

    def __init__(self, values):
        self.values = values

    def get_contents(self):
        return ""

    @classmethod
    def from_info(cls, project_info):
        elements = getattr(cls, "elements", [])
        values = {}
        for element in elements:
            values[element.name] = element.get_value(project_info)
        return cls(values)


class TemplatePackageFile(PackageFile):

    def get_contents(self):
        t = TemplateFile(self.template)
        return t.render(self.values)


class DebianPackageFile(PackageFile):

    @property
    def path(self):
        return os.path.join(DEBIAN_DIR, self.filename)


class PackageFileGroup(object):

    def __init__(self):
        self.files_cls = []

    def add_file_cls(self, package_file_cls):
        self.files_cls.append(package_file_cls)

    def get_files(self, project_info):
        elements = self.get_elements()
        # XXX: Kind of hacky.  The thing is that the ExtraFiles element isn't
        # an informational element used by key files, but rather a way of
        # forcing extra files to be written.
        keys = [ExtraFiles.name] + [element.name for element in elements]
        values = project_info.get_all(keys)
        new_info = DictInfo(values)
        files = []
        for file_cls in self.files_cls:
            files.append(file_cls.from_info(new_info))
        extra_files = ExtraFiles.get_value(project_info)
        for path, contents in extra_files.items():
            files.append(BasicFile(path, contents))
        return files

    def get_elements(self):
        elements = set()
        for file_cls in self.files_cls:
            file_elements = getattr(file_cls, "elements", [])
            for element in file_elements:
                elements.add(element)
        return elements


default_package_file_group = PackageFileGroup()


class Compat(DebianPackageFile):

    filename = "compat"

    def get_contents(self):
        return "7\n"


default_package_file_group.add_file_cls(Compat)


class Rules(DebianPackageFile, TemplatePackageFile):

    elements = [
        Buildsystem,
        DebhelperAddons,
        ]

    filename = "rules"
    template = "rules"


default_package_file_group.add_file_cls(Rules)


class Copyright(DebianPackageFile):

    filename = "copyright"


default_package_file_group.add_file_cls(Copyright)


class Changelog(DebianPackageFile):

    elements = [
        PackageName,
        Version,
        Maintainer,
        ]

    filename = "changelog"

    def get_contents(self):
        cl = changelog.Changelog()
        cl.new_block(
            package=self.values[PackageName.name],
            version=self.values[Version.name],
            distributions="unstable",
            changes=["", "  * New version", ""],
            author=self.values[Maintainer.name],
            date=datetime.datetime.utcnow().strftime(
                "%a, %d %b %Y %H:%M:%S +0000"),
            urgency="low",
            )
        return str(cl)


default_package_file_group.add_file_cls(Changelog)


class Control(DebianPackageFile, TemplatePackageFile):

    elements = [
        Architecture,
        BuildDepends,
        Depends,
        Description,
        Homepage,
        Maintainer,
        PackageName,
        Section,
        ]

    filename = "control"
    template = "control"


default_package_file_group.add_file_cls(Control)


class Desktop(TemplatePackageFile):

    # See http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s05.html

    elements = [
        ApplicationName,
        Categories,
        Executable,
        PackageName,
        TagLine,
        ]

    template = 'desktop'

    @property
    def path(self):
        return '%s.desktop' % (self.values[PackageName.name],)
