import errno
import os
import subprocess


class ScriptProblem(Exception):

    def __init__(self, script_path):
        self.script_path = script_path


class ScriptMissing(ScriptProblem):

    def __init__(self, script_path):
        super(ScriptMissing, self).__init__(script_path)

    def __str__(self):
        return "%s not found" % (self.script_path, )


class ScriptFailed(ScriptProblem):

    def __init__(self, script_path, returncode, output):
        super(ScriptFailed, self).__init__(script_path)
        self.returncode = returncode
        self.output = output

    def __str__(self):
        return (
            "%s failed with returncode %d, output was %s"
            % (self.script_path, self.returncode, self.output))


def run_script(basepath, script_name, cwd, to_write=None):
    script_path = os.path.join(os.path.abspath(basepath), script_name)
    helperpath = os.path.join(
        os.path.dirname(os.path.dirname(basepath)), 'helpers')
    env = os.environ.copy()
    if os.path.isdir(helperpath):
        if 'PATH' in env:
            env['PATH'] = os.pathsep.join([env['PATH'], helperpath])
        else:
            env['PATH'] = helperpath
    try:
        proc = subprocess.Popen(
            [script_path], stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, stdin=subprocess.PIPE,
            cwd=cwd, env=env)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ScriptMissing(script_path)
        raise
    out, ignore = proc.communicate(to_write)
    if proc.returncode != 0:
        raise ScriptFailed(script_path, proc.returncode, out)
    return out
