import os
import sys

from fixtures import TestWithFixtures
from testtools import TestCase

from pkgme.backend import (
    BackendSelector,
    EXTERNAL_BACKEND_PATHS,
    ExternalHelpersBackend,
    ExternalHelpersBackendLoader,
    get_default_loader,
    get_default_selector,
    get_info_for,
    NoBackend,
    reset_default_loader_factory,
    set_default_loader_factory,
    StaticBackend,
    StaticLoader,
    )
from pkgme.project_info import (
    MultipleExternalHelpersInfo,
    SingleExternalHelperInfo,
    )
from pkgme.testing import (
    ExecutableFileFixture,
    TempdirFixture,
    )


class ExternalHelpersBackendTests(TestCase, TestWithFixtures):

    def make_want_script(self, script):
        tempdir = self.useFixture(TempdirFixture())
        script_name = ExternalHelpersBackend.WANT_SCRIPT_NAME
        script_path = os.path.join(tempdir.path, script_name)
        self.useFixture(ExecutableFileFixture(script_path, script))
        return tempdir

    def test_want_runs_script(self):
        script = "#!/bin/sh\necho 10\n"
        tempdir = self.make_want_script(script)
        backend = ExternalHelpersBackend(self.getUniqueString(), tempdir.path)
        self.assertEqual(10, backend.want(tempdir.path))

    def test_missing_want_script(self):
        tempdir = self.useFixture(TempdirFixture())
        name = self.getUniqueString()
        backend = ExternalHelpersBackend(name, tempdir.path)
        e = self.assertRaises(AssertionError, backend.want, tempdir.path)
        self.assertEqual(
            "Backend %s (%s) has no '%s' script"
            % (name, tempdir.path, ExternalHelpersBackend.WANT_SCRIPT_NAME),
            str(e))

    def test_want_script_returns_non_integer(self):
        script = "#!/bin/sh\necho foo\n"
        tempdir = self.make_want_script(script)
        name = self.getUniqueString()
        backend = ExternalHelpersBackend(name, tempdir.path)
        e = self.assertRaises(AssertionError, backend.want, tempdir.path)
        self.assertEqual(
            "Backend %s (%s) returned non-integer score from '%s' script: foo"
            % (name, tempdir.path, ExternalHelpersBackend.WANT_SCRIPT_NAME),
            str(e))

    def test_want_script_run_in_correct_dir(self):
        cwd_tempdir = self.useFixture(TempdirFixture())
        script = """#!%s
import os
if os.getcwd() != '%s':
    print "Backend was called in %%s rather than %%s" %% (os.getcwd(), '%s')
else:
    print 10
""" % (sys.executable, cwd_tempdir.path, cwd_tempdir.path)
        backend_tempdir = self.make_want_script(script)
        backend = ExternalHelpersBackend(
            self.getUniqueString(), backend_tempdir.path)
        self.assertEqual(10, backend.want(cwd_tempdir.path))

    def test_get_info_with_multiple(self):
        helpers_path = self.getUniqueString()
        cwd_path = self.getUniqueString()
        backend = ExternalHelpersBackend(self.getUniqueString(), helpers_path)
        project_info = backend.get_info(cwd_path)
        self.assertIsInstance(project_info, MultipleExternalHelpersInfo)
        self.assertEqual(helpers_path, project_info.basepath)
        self.assertEqual(cwd_path, project_info.cwd)

    def test_get_info_with_single(self):
        helpers_tempdir = self.useFixture(TempdirFixture())
        info_script_path = os.path.join(
            helpers_tempdir.path,
            SingleExternalHelperInfo.INFO_SCRIPT_NAME)
        with open(info_script_path, "w"):
            pass
        cwd_path = self.getUniqueString()
        backend = ExternalHelpersBackend(
            self.getUniqueString(), helpers_tempdir.path)
        project_info = backend.get_info(cwd_path)
        self.assertIsInstance(project_info, SingleExternalHelperInfo)
        self.assertEqual(helpers_tempdir.path, project_info.basepath)
        self.assertEqual(cwd_path, project_info.cwd)


class BackendSelectorTests(TestCase):

    def test_selects_highest_score(self):
        info1 = self.getUniqueString()
        info2 = self.getUniqueString()
        path = self.getUniqueString()
        backend1 = StaticBackend(
            self.getUniqueString(), 10, info1, expected_path=path)
        backend2 = StaticBackend(
            self.getUniqueString(), 20, info2, expected_path=path)
        backends = [backend1, backend2]
        selector = BackendSelector(backends)
        self.assertEqual(info2, selector.get_info(path))

    def test_no_backends(self):
        path = self.getUniqueString()
        selector = BackendSelector([])
        self.assertRaises(NoBackend, selector.get_info, path)

    def test_no_eligble(self):
        info = self.getUniqueString()
        path = self.getUniqueString()
        backend = StaticBackend(
            self.getUniqueString(), 0, info, expected_path=path)
        backends = [backend]
        selector = BackendSelector(backends)
        self.assertRaises(NoBackend, selector.get_info, path)

    def test_selects_first_name_in_tie(self):
        info1 = self.getUniqueString()
        info2 = self.getUniqueString()
        path = self.getUniqueString()
        backend1 = StaticBackend("a", 10, info1, expected_path=path)
        backend2 = StaticBackend("b", 10, info2, expected_path=path)
        backends = [backend1, backend2]
        selector = BackendSelector(backends)
        self.assertEqual(info1, selector.get_info(path))


class ExternalHelpersBackendLoaderTests(TestCase, TestWithFixtures):

    def test_loads_no_backends(self):
        tempdir = self.useFixture(TempdirFixture())
        loader = ExternalHelpersBackendLoader([tempdir.path])
        self.assertEqual([], loader.load())

    def test_loads_from_one_dir(self):
        tempdir = self.useFixture(TempdirFixture())
        backend_name = self.getUniqueString()
        basepath = tempdir.mkdir(backend_name)
        loader = ExternalHelpersBackendLoader([tempdir.path])
        backends = loader.load()
        self.assertEqual(1, len(backends))
        self.assertIsInstance(backends[0], ExternalHelpersBackend)
        self.assertEqual(backend_name, backends[0].name)
        self.assertEqual(basepath, backends[0].basepath)

    def test_loads_from_two_dirs(self):
        tempdir1 = self.useFixture(TempdirFixture())
        tempdir2 = self.useFixture(TempdirFixture())
        backend_name1 = self.getUniqueString()
        backend_name2 = self.getUniqueString()
        tempdir1.mkdir(backend_name1)
        tempdir2.mkdir(backend_name2)
        loader = ExternalHelpersBackendLoader(
            [tempdir1.path, tempdir2.path])
        backends = loader.load()
        self.assertEqual(2, len(backends))

    def test_loads_only_one_instance_of_each_name(self):
        tempdir1 = self.useFixture(TempdirFixture())
        tempdir2 = self.useFixture(TempdirFixture())
        backend_name = self.getUniqueString()
        basepath1 = tempdir1.mkdir(backend_name)
        tempdir2.mkdir(backend_name)
        loader = ExternalHelpersBackendLoader(
            [tempdir1.path, tempdir2.path])
        backends = loader.load()
        self.assertEqual(1, len(backends))
        self.assertEqual(basepath1, backends[0].basepath)

    def test_loads_only_dirs(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch(self.getUniqueString())
        loader = ExternalHelpersBackendLoader([tempdir.path])
        self.assertEqual([], loader.load())

    def test_ignores_dotdirs(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.mkdir("." + self.getUniqueString())
        loader = ExternalHelpersBackendLoader([tempdir.path])
        self.assertEqual([], loader.load())

    def test_ignores_missing_dirs(self):
        self.useFixture(TempdirFixture())
        loader = ExternalHelpersBackendLoader([self.getUniqueString()])
        self.assertEqual([], loader.load())


class GetDefaultSelectorTests(TestCase):

    def test_returns_backend_selector(self):
        backend = self.getUniqueString()
        selector = get_default_selector([backend])
        self.assertIsInstance(selector, BackendSelector)
        self.assertEqual([backend], selector.backends)


class GetDefaultLoaderTests(TestCase):

    def test_returns_external_helpers_loader(self):
        loader = get_default_loader()
        self.assertIsInstance(loader, ExternalHelpersBackendLoader)

    def test_sets_default_paths(self):
        loader = get_default_loader()
        self.assertEqual(EXTERNAL_BACKEND_PATHS, loader.paths)


class TestSelector(BackendSelector):

    def get_info(self, path):
        return self.__class__.__name__


class GetInfoForTests(TestCase):

    def test_gets_correct_info(self):
        info = self.getUniqueString()
        path = self.getUniqueString()
        backend = StaticBackend(
            self.getUniqueString(), 10, info, expected_path=path)
        loader = StaticLoader([backend])
        self.assertEqual(info, get_info_for(path, loader=loader))

    def test_uses_selector_cls(self):
        info = self.getUniqueString()
        path = self.getUniqueString()
        backend = StaticBackend(
            self.getUniqueString(), 10, info, expected_path=path)
        loader = StaticLoader([backend])
        selector_cls = TestSelector
        self.assertEqual(
            selector_cls.__name__,
            get_info_for(path, loader=loader, selector_cls=selector_cls))

    def test_uses_default_loader(self):
        info = self.getUniqueString()
        path = self.getUniqueString()
        backend = StaticBackend(
            self.getUniqueString(), 10, info, expected_path=path)
        def make_loader():
            return StaticLoader([backend])
        self.addCleanup(reset_default_loader_factory)
        set_default_loader_factory(make_loader)
        self.assertEqual(info, get_info_for(path))
