import os

from fixtures import TestWithFixtures
from testtools import TestCase
from testtools.matchers import StartsWith

from pkgme.binary import (
    apt_file,
    find_shared_library_dependencies,
    get_desktop_file,
    get_file_type,
    get_file_types,
    get_install_file,
    get_shared_library_dependencies,
    guess_dependencies,
    guess_executable,
    iter_binaries,
    iter_executables,
    ldd,
    METADATA_FILE,
    )
from pkgme.package_files import DEBIAN_DIR
from pkgme.testing import (
    AreDesktopValuesFor,
    TempdirFixture,
    )


class BinaryTests(TestCase, TestWithFixtures):

    def test_metadata_file(self):
        self.assertEqual('devportal-metadata.json', METADATA_FILE)


class InstallFileTests(TestCase, TestWithFixtures):

    def test_install_file(self):
        # The install file instructs debhelper to copy everything in the
        # top-level to /opt/<package-name>.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file')
        install_file = get_install_file('package-name', tempdir.path)
        self.assertEqual("some-file opt/package-name\n", install_file)

    def test_install_file_many_files_and_directories(self):
        # The install file instructs debhelper to copy everything in the
        # top-level to /opt/<package-name>.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file')
        tempdir.mkdir('directory')
        install_file = get_install_file('package-name', tempdir.path)
        self.assertEqual(
            "directory opt/package-name\n"
            "some-file opt/package-name\n",
            install_file)

    def test_skip_debian(self):
        # The install file instructs debhelper to copy everything in the
        # top-level to /opt/<package-name>, except for the 'debian' directory.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file')
        tempdir.mkdir('directory')
        tempdir.mkdir(DEBIAN_DIR)
        install_file = get_install_file('package-name', tempdir.path)
        self.assertEqual(
            "directory opt/package-name\n"
            "some-file opt/package-name\n",
            install_file)

    def test_skip_metadata(self):
        # The install file instructs debhelper to copy everything in the
        # top-level to /opt/<package-name>, except for the 'debian' directory
        # and the metadata file.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file')
        tempdir.mkdir('directory')
        tempdir.touch(METADATA_FILE)
        install_file = get_install_file('package-name', tempdir.path)
        self.assertEqual(
            "directory opt/package-name\n"
            "some-file opt/package-name\n",
            install_file)

    def test_include_desktop_file(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file')
        install_file = get_install_file(
            'package-name', tempdir.path, include_desktop=True)
        self.assertEqual(
            "some-file opt/package-name\n"
            "debian/package-name.desktop usr/share/application\n",
            install_file)


class DesktopFileTests(TestCase):

    def test_executable_is_best_guess(self):
        package_name = self.getUniqueString()
        tempdir = self.useFixture(TempdirFixture())
        tempdir.mkdir('whatever')
        tempdir.touch('whatever/not-the-best', 0755)
        tempdir.touch('the-best', 0755)
        desktop_file = get_desktop_file(package_name, tempdir.path)
        self.assertThat(
            {'Exec': '/opt/%s/the-best' % (package_name,)},
            AreDesktopValuesFor(desktop_file))

    def test_app_name_is_capitalized_package_name(self):
        # We don't have any information on the package name, so try to guess
        # the application name.
        tempdir = self.useFixture(TempdirFixture())
        package_name = self.getUniqueString()
        desktop_file = get_desktop_file(package_name, tempdir.path)
        self.assertThat(
            {'Name': package_name.capitalize()},
            AreDesktopValuesFor(desktop_file))

    def test_category_and_tagline_are_specified(self):
        # We just pass the category and comment through.
        tempdir = self.useFixture(TempdirFixture())
        package_name = self.getUniqueString()
        tagline = self.getUniqueString()
        categories = self.getUniqueString()
        desktop_file = get_desktop_file(
            package_name, tempdir.path, tagline=tagline, categories=categories)
        self.assertThat(
            {'Comment': tagline, 'Categories': categories},
            AreDesktopValuesFor(desktop_file))


class FindExecutableTests(TestCase, TestWithFixtures):

    def test_only_one_file_and_it_is_executable(self):
        # If there is only one file and it's executable, find that.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file', mode=0755)
        executables = list(iter_executables(tempdir.path))
        self.assertEqual(['some-file'], executables)

    def test_no_files_at_all(self):
        # iter_executables finds no executables if there are no files at all.
        tempdir = self.useFixture(TempdirFixture())
        executables = list(iter_executables(tempdir.path))
        self.assertEqual([], executables)

    def test_no_executable_files(self):
        # If there are no executable files, iter_executables returns None.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file', mode=0644)
        executables = list(iter_executables(tempdir.path))
        self.assertEqual([], executables)

    def test_directory_is_not_executable_file(self):
        # A directory does not count as an executable file.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.mkdir('directory')
        executables = list(iter_executables(tempdir.path))
        self.assertEqual([], executables)

    def test_finds_executable_in_nested_directory(self):
        # Even if the file is in some nested directory, we are able to find
        # it.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.mkdir('directory')
        tempdir.touch('directory/my-executable', mode=0755)
        executables = list(iter_executables(tempdir.path))
        self.assertEqual(['directory/my-executable'], executables)

    def test_multiple_executables(self):
        # If there are many executables, iter_executables finds them all.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch('some-file', mode=0755)
        tempdir.touch('another-file', mode=0755)
        executables = sorted(list(iter_executables(tempdir.path)))
        self.assertEqual(['another-file', 'some-file'], executables)


class GuessExecutableTests(TestCase):

    def test_no_executables(self):
        # If there are no executables to select from, then return None,
        # indicating the fact.
        executable = guess_executable('package-name', iter([]))
        self.assertIs(None, executable)

    def test_only_one_executable(self):
        # If there's only one executable, then return that, since it's
        # probably the main executable for the package.
        executable = guess_executable('package-name', ['whatever'])
        self.assertEqual('whatever', executable)

    def test_exact_package_name_match(self):
        # If there are many executables, but one of them has the same name as
        # the package, then that is probably the main executable.
        executable = guess_executable(
            'package-name', ['whatever', 'package-name'])
        self.assertEqual('package-name', executable)

    def test_exact_package_name_match_in_subdir(self):
        # If there are many executables, but one of them has the same name as
        # the package, then that is probably the main executable, even if it
        # is in a sub-directory.
        executable = guess_executable(
            'package-name', ['whatever', 'subdir/package-name', 'foo'])
        self.assertEqual('subdir/package-name', executable)

    def test_multiple_exact_matches(self):
        # If there are many executables that have the same name as the
        # package, then the one that is the least nested is our best guess.
        executable = guess_executable(
            'package-name', [
                'whatever', 'a/b/c/d/e/subdir/package-name', 'foo',
                'subdir/package-name'])
        self.assertEqual('subdir/package-name', executable)

    def test_different_case_match(self):
        # If one of the executables has the same name as the package, but
        # spelled with different case, then that is our best guess.
        executable = guess_executable(
            'packagename', [
                'whatever', 'a/b/c/d/e/subdir/packagename', 'foo',
                'subdir/PackageName'])
        self.assertEqual('subdir/PackageName', executable)

    def test_many_executables(self):
        # If there are many executables, and their names have no particular
        # relationship to the package name, then just pick the top-most
        # one. If there's more than one, take the alphabetically sorted.
        executable = guess_executable(
            'package-name', ['dir/x', 'dir/sub/y', 'z', 'a'])
        self.assertEqual('a', executable)


class GetFileTypeTests(TestCase):

    def test_plain_text(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.create_file('foo.txt', 'boring content\n')
        file_type = get_file_type(tempdir.abspath('foo.txt'))
        self.assertEqual('ASCII text', file_type)

    def test_data(self):
        file_type = get_file_type(
            os.path.join(os.path.dirname(__file__), 'foo.data'))
        self.assertEqual('data', file_type)

    def test_elf_binary(self):
        binary_path = os.path.join(os.path.dirname(__file__), 'hello')
        file_type = get_file_type(binary_path)
        self.assertThat(file_type, StartsWith('ELF'))

    def test_elf_library(self):
        binary_path = os.path.join(os.path.dirname(__file__), 'simple.so.1')
        file_type = get_file_type(binary_path)
        self.assertThat(file_type, StartsWith('ELF'))

    def test_multiple(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.create_file('foo.txt', 'boring content\n')
        file_types = get_file_types(
            [tempdir.abspath('foo.txt'),
             os.path.join(os.path.dirname(__file__), 'foo.data')])
        self.assertEqual(['ASCII text', 'data'], file_types)

    def test_no_files_given(self):
        self.assertEqual([], get_file_types([]))
        self.assertEqual([], get_file_types(iter([])))


class IterBinariesTests(TestCase):

    def test_no_binaries(self):
        tempdir = self.useFixture(TempdirFixture())
        self.assertEqual([], list(iter_binaries(tempdir.path)))

    def test_some_binaries(self):
        path = os.path.dirname(os.path.dirname(__file__))
        binaries = sorted(iter_binaries(path))
        self.assertEqual(
            [os.path.join(os.path.dirname(__file__), 'hello'),
             os.path.join(os.path.dirname(__file__), 'simple.so.1')],
            binaries)


class GetSharedLibraryDependenciesTests(TestCase):

    def test_ldd_none(self):
        deps = ldd([])
        self.assertEqual({}, deps)

    def test_ldd(self):
        hello = os.path.join(os.path.dirname(__file__), 'hello')
        deps = ldd([hello])
        self.assertEqual(
            {hello: [(None, '/lib64/ld-linux-x86-64.so.2'),
                     ('libc.so.6', '/lib/x86_64-linux-gnu/libc.so.6'),
                     ('linux-vdso.so.1', None)]},
            deps)

    def test_ldd_multiple(self):
        hello = os.path.join(os.path.dirname(__file__), 'hello')
        simple = os.path.join(os.path.dirname(__file__), 'simple.so.1')
        deps = ldd([hello, simple])
        self.assertEqual(
            {hello: [(None, '/lib64/ld-linux-x86-64.so.2'),
                     ('libc.so.6', '/lib/x86_64-linux-gnu/libc.so.6'),
                     ('linux-vdso.so.1', None)],
             simple: [(None, '/lib64/ld-linux-x86-64.so.2'),
                     ('libc.so.6', '/lib/x86_64-linux-gnu/libc.so.6'),
                     ('linux-vdso.so.1', None)]},
            deps)

    def test_get_shared_library_dependencies(self):
        hello = os.path.join(os.path.dirname(__file__), 'hello')
        deps = sorted(get_shared_library_dependencies([hello]))
        self.assertEqual(
            ['/lib/x86_64-linux-gnu/libc.so.6',
             '/lib64/ld-linux-x86-64.so.2',
             ], deps)

    def test_find_shared_library_dependencies(self):
        deps = find_shared_library_dependencies(os.path.dirname(__file__))
        self.assertEqual(
            set(['/lib/x86_64-linux-gnu/libc.so.6',
                 '/lib64/ld-linux-x86-64.so.2']), deps)


class AptFileTests(TestCase):

    def test_full_file(self):
        packages = list(apt_file('/usr/lib32/libasound.so.2'))
        self.assertEqual(['lib32asound2'], packages)


class GuessDependenciesTests(TestCase):

    def test_guess_dependencies(self):
        deps = guess_dependencies(os.path.dirname(__file__))
        self.assertEqual(set(['libc6']), deps)
