import json
import os
import shutil

from fixtures import TestWithFixtures
from testtools import TestCase

from pkgme.backend import ExternalHelpersBackend, get_backend_dir
from pkgme.binary import (
    CATEGORIES,
    DEPENDS,
    get_desktop_file,
    get_install_file,
    guess_dependencies,
    METADATA_FILE,
    PACKAGE_NAME,
    TAGLINE,
    )
from pkgme.info_elements import (
    BuildDepends,
    Depends,
    ExtraFiles,
    PackageName,
    )
from pkgme.testing import TempdirFixture


BACKEND_NAME = 'binary'

backend_dir = get_backend_dir(__file__, BACKEND_NAME)


class BinaryBackendTests(TestCase, TestWithFixtures):

    def test_want_with_metadata(self):
        # If we detect a binary, then we score 10. The way we determine if
        # something is a binary is if it has a devportal-metadata.json in its
        # top-level.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.touch(METADATA_FILE)
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        self.assertEqual(10, backend.want(tempdir.path))

    def test_want_without_metadata(self):
        # If we do *not* detect a binary, then we score 0. The way we
        # determine if something is a binary is if it has a
        # devportal-metadata.json in its top-level.
        tempdir = self.useFixture(TempdirFixture())
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        self.assertEqual(0, backend.want(tempdir.path))

    def test_package_name(self):
        # The binary backend gets the package name from the metadata file.
        tempdir = self.useFixture(TempdirFixture())
        tempdir.create_file(METADATA_FILE, json.dumps({PACKAGE_NAME: 'foo'}))
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        info = backend.get_info(tempdir.path)
        self.assertEqual(
            {PackageName.name: "foo"}, info.get_all([PackageName.name]))

    def test_build_depends(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.create_file(METADATA_FILE, json.dumps({PACKAGE_NAME: 'foo'}))
        shutil.copy(
            os.path.join(os.path.dirname(__file__), 'hello'),
            tempdir.path)
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        info = backend.get_info(tempdir.path)
        deps = ', '.join(guess_dependencies(tempdir.path))
        self.assertEqual(
            {BuildDepends.name: deps}, info.get_all([BuildDepends.name]))

    def test_depends(self):
        tempdir = self.useFixture(TempdirFixture())
        tempdir.create_file(METADATA_FILE, json.dumps({PACKAGE_NAME: 'foo'}))
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        info = backend.get_info(tempdir.path)
        self.assertEqual({Depends.name: DEPENDS}, info.get_all([Depends.name]))

    def test_extra_files_install_file(self):
        # We create an 'install' file that tells debhelper to just copy
        # everything to opt.
        tempdir = self.useFixture(TempdirFixture())
        package_name = self.getUniqueString()
        tempdir.create_file(METADATA_FILE, json.dumps({PACKAGE_NAME: package_name}))
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        info = backend.get_info(tempdir.path)
        extra_files = info.get_all([ExtraFiles.name])[ExtraFiles.name]
        install_file = json.loads(extra_files)['debian/install']
        self.assertEqual(
            get_install_file(package_name, tempdir.path, True),
            install_file)

    def test_extra_files_desktop_file(self):
        # We create an 'install' file that tells debhelper to just copy
        # everything to opt.
        tempdir = self.useFixture(TempdirFixture())
        package_name = self.getUniqueString()
        tagline = self.getUniqueString()
        categories = self.getUniqueString()
        metadata = {
            PACKAGE_NAME: package_name,
            TAGLINE: tagline,
            CATEGORIES: categories,
            }
        tempdir.create_file(METADATA_FILE, json.dumps(metadata))
        # We need to create an executable in order to be able to generate a
        # desktop.
        tempdir.touch('executable', 0755)
        expected_desktop_file = get_desktop_file(
            package_name, tempdir.path, tagline=tagline,
            categories=categories).get_contents()
        backend = ExternalHelpersBackend(BACKEND_NAME, backend_dir)
        info = backend.get_info(tempdir.path)
        extra_files = info.get_all([ExtraFiles.name])[ExtraFiles.name]
        desktop = json.loads(extra_files)['debian/%s.desktop' % (package_name,)]
        self.assertEqual(expected_desktop_file, desktop)
