import json

from testtools import TestCase

from pkgme.info_elements import (
    BuildDepends,
    ExtraFiles,
    InfoElement,
    MissingInfoError,
    )
from pkgme.project_info import DictInfo


class TestElement(InfoElement):

    name = "test_element"


class RequiredElement(InfoElement):

    name = "required_element"
    required = True
    default = "foo"


class DefaultElement(InfoElement):

    name = "default_element"
    default = "bar"


class TestException(Exception):
    pass


class AlwaysInvalidElement(InfoElement):

    name = "always_invalid_element"

    @classmethod
    def clean(cls, value):
        raise TestException()


class CleaningElement(InfoElement):

    name = "cleaning_element"
    return_value = "some_return_value"

    @classmethod
    def clean(cls, value):
        return cls.return_value


class InfoElementTests(TestCase):

    def test_get_value(self):
        value = self.getUniqueString()
        project_info = DictInfo({TestElement.name: value})
        self.assertEqual(value, TestElement.get_value(project_info))

    def test_get_value_without_value(self):
        self.assertEqual(None, TestElement.get_value(DictInfo({})))

    def test_get_value_with_default(self):
        self.assertEqual("bar", DefaultElement.get_value(DictInfo({})))

    def test_value_required(self):
        e = self.assertRaises(
            MissingInfoError, RequiredElement.get_value, DictInfo({}))
        self.assertEqual(
            "The backend didn't provide a value for 'required_element'",
            str(e))

    def test_clean_can_raise_exception(self):
        self.assertRaises(
            TestException, AlwaysInvalidElement.get_value, DictInfo({}))

    def test_uses_cleaned_value(self):
        original_value = self.getUniqueString()
        self.assertNotEqual(CleaningElement.return_value, original_value)
        project_info = DictInfo({CleaningElement.name: original_value})
        self.assertEqual(
            CleaningElement.return_value, CleaningElement.get_value(project_info))

    def test_default_element_doesnt_clean_value(self):
        original_value = self.getUniqueString()
        project_info = DictInfo({TestElement.name: original_value})
        self.assertEqual(original_value, TestElement.get_value(project_info))


class BuildDependsTestCase(TestCase):

    def test_clean_inserts_debhelper(self):
        value = BuildDepends.clean(None)
        self.assertEqual("debhelper (>= 7)", value)

    def test_clean_adds_debhelper(self):
        value = BuildDepends.clean("python")
        self.assertEqual("python, debhelper (>= 7)", value)

    def test_clean_doesnt_add_debhelper_if_present(self):
        value = BuildDepends.clean("python, debhelper")
        self.assertEqual("python, debhelper", value)

    def test_clean_converts_list(self):
        # BuildDepends can be specified as a list via JSON output from
        # `all_info`.
        value = BuildDepends.clean(["python", "gtk2"])
        self.assertEqual("python, gtk2, debhelper (>= 7)", value)


class ExtraFilesTestCase(TestCase):

    def test_clean_preserves_dict(self):
        in_value = {'foo': 'bar'}
        out_value = ExtraFiles.clean(in_value)
        self.assertEqual(out_value, in_value)

    def test_clean_loads_json(self):
        in_value = {'foo': 'bar'}
        out_value = ExtraFiles.clean(json.dumps(in_value))
        self.assertEqual(out_value, in_value)
